/*
  RabidVWM LiteStep Module - version 1.0
	Window List Management (low level)

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*

Windows are stored in an array, sorted by window handle.  This way
given a handle, a quick binary search can be used to find an existing
window, or the location to insert a new window.

The array has a fairly large initial size, but in case someone actually
opens enough windows to overflow it, InsertWindow will automatically
increase the space if needed.  The extra space is not released until
the module exits.  (I assume that anyone openning this many windows will
be using a lot of windows for a while, and it's not a *lot* of memory.)

*/

#include "RabidVWM.h"

// allocate or reallocate memory for window list
void AllocateWindows(int nSpace)
{
	s_WinData *pData;
	pData = GlobalData.KnownWindows;
	TRACE(TRACE_STATUS,"Allocating KnownWindows.");
	if (nSpace<GlobalData.nKnownWindows) {
		TRACE(TRACE_ERROR,"Failed - nSpace too small");
		return;
	}
	GlobalData.KnownWindows = new s_WinData[nSpace];
	/* if it fails, keep the old list           */ 
	if (!GlobalData.KnownWindows) {
		TRACE(TRACE_WARNING,"failed.");
		GlobalData.KnownWindows = pData;
		return;
	}
	/* copy any old window data to the new list */ 
	if (pData && GlobalData.nKnownWindows) {
		CopyMemory(
			GlobalData.KnownWindows,
			pData,
			GlobalData.nKnownWindows*sizeof(s_WinData));
		delete[] pData;
	}
	GlobalData.nMaxWindows = nSpace;
}

// Insert a window into the known window list at a given position,
// it does not preserve sort order.
// This (used to be) static because it's easy to screw up the list with it.
// Use FindInsertWindow instead.
int InsertWindow(int position,HWND hWnd)
{
	int i;
	s_WinData *pData;

	ASSERT(position>=0);
	ASSERT(position<=GlobalData.nKnownWindows);
	ASSERT(hWnd!=NULL);

	/* grow space if there isn't enough   */ 
	if (GlobalData.nKnownWindows+1>=GlobalData.nMaxWindows)
		AllocateWindows(GlobalData.nMaxWindows+10);

	pData = &GlobalData.KnownWindows[position];

	/* don't add it if it's already there */ 
	if (hWnd==pData->hWnd)
		return position;

	if (position<GlobalData.nKnownWindows)
		if (pData->hWnd<hWnd) { position++; pData++; }

	/* pRemember needs to remain valid    */ 
	if (GlobalData.pRemember>=pData)
		GlobalData.pRemember++;

	/* push the rest forward to make room */ 
	for (i=GlobalData.nKnownWindows;i>position;--i)
		GlobalData.KnownWindows[i] = GlobalData.KnownWindows[i-1];

	/* clear the space                    */ 
	ZeroMemory(pData,sizeof(s_WinData));
	pData->hWnd = hWnd;
	pData->bTouch = GlobalData.bWindowTouch;
	pData->bNewWindow = true;
	pData->HomeDesk = -1;
	pData->Alpha = GlobalData.cWnd.Alpha;
	GlobalData.nKnownWindows++;

#ifndef NDEBUG
	/* make sure the list is in order     */ 
	for (i=1;i<GlobalData.nKnownWindows;i++)
		ASSERT(GlobalData.KnownWindows[i-1].hWnd<GlobalData.KnownWindows[i].hWnd);
#endif

	return position;
}

// remove the window at the given position
void DeleteWindow(int position)
{
	int i;
	s_WinData *pData;

	ASSERT(position>=0);
	ASSERT(position<GlobalData.nKnownWindows);

	pData = &GlobalData.KnownWindows[position];
	/* pRemember needs to remain valid    */ 
	if (GlobalData.pRemember>pData)
		/* if it's equal, you have problems */ 
		GlobalData.pRemember--;

  /* if we're deleting the pRemember window, perhaps it should */ 
  /* be set to NULL so that whoever's using it (dragging code) */ 
  /* can know to cancel the operation...                       */ 
  /* Not setting to NULL until I check that it will cancel.    */ 

	/* free things allocated to window    */ 
	if (pData->hSnapShot) {
		DeleteObject(pData->hSnapShot);
		StatGDIDel(Bitmap);
		pData->hSnapShot = NULL; /* used as flag for delete */ 
	}

	for (i=position;i<GlobalData.nKnownWindows;i++)
		GlobalData.KnownWindows[i] = GlobalData.KnownWindows[i+1];

	GlobalData.nKnownWindows--;
}

// binary search through KnownWindows
// returns index of closest window handle found
int FindKnownWindow(HWND hWnd)
{
	int l,h,m;

	ASSERT(GlobalData.KnownWindows!=NULL);

	l = 0; h = GlobalData.nKnownWindows-1;

	/* no known windows  */ 
	if (h<l) return l;

	/* check edges first */ 
	if (hWnd<=GlobalData.KnownWindows[l].hWnd) return l;
	if (hWnd>=GlobalData.KnownWindows[h].hWnd) return h;

	/* binary search     */ 
	do {
		m = (h+l)/2;
		if (hWnd==GlobalData.KnownWindows[m].hWnd) return m;
		if (hWnd< GlobalData.KnownWindows[m].hWnd) h=m; else l=m;
	} while (h-l>1);

	ASSERT(m>=0);
	ASSERT(m<GlobalData.nKnownWindows);

	return m;
}

// will never return NULL, always check for pData->hWnd==hWnd
s_WinData *GetWindowPtr(HWND hWnd)
{
	int i;
  s_WinData *pData;

	ASSERT(GlobalData.KnownWindows!=NULL);

	i = FindKnownWindow(hWnd);
  pData = &GlobalData.KnownWindows[i];

  ASSERT(READABLE(pData,sizeof(*pData)));
  ASSERT(WRITEABLE(pData,sizeof(*pData)));
  return pData;
}

// binary search through KnownWindows
// return index in KnownWindows, window is inserted if not found
int FindInsertWindow(HWND hWnd)
{
	int m = FindKnownWindow(hWnd);
	/* m = position of item if present,   */ 
  /*     or position where it should go */ 
	return InsertWindow(m,hWnd);
}
